#!/usr/bin/python
# -*- coding: utf-8 -*-

# parser for noreo.com's ocean girl board (python 2.6). some tweaks and it's
# /FAST/ now, heya!
# public domain (use it for any purpose)
#
# usage:
# $ find data_index/ data/ -type f | ./convert.py banned.list.txt > all_filtered.html
#
# note: ids > 1.000.000 have been manually assigned due to conflicts

import re, sys, os
from datetime import datetime, timedelta

def multireplace (s, replace):
	for searchitem, replaceitem in replace:
		s = s.replace (searchitem, replaceitem)
	return s

def entityDecode (txt):
	return multireplace (txt, [('&quot;', '"'), ('&amp;', '&'), ('&copy;', u'©'), ('&lt;', '<'), ('&gt;', '>')])

class MustNotHappen (Exception):
	""" should not happen. don't catch this exception, fix the code instead """
	pass

class Counter(object):
	def __init__ (self, name):
		self.name = name
		self.started = datetime.now ()
	
	def __del__ (self):
		sys.stderr.write ('counter ' + self.name + ' stopped: ' + unicode (datetime.now () - self.started) + '\n')

class Post(object):
	NOPARENT = 0

	# for fromHtml()
	TITLERE = re.compile (r'<title>(.*?)</title>')
	CONTENTRE = re.compile (r'</font><hr width=100%><br>(.*?)<a name="followups">', re.DOTALL)
	TIMEAUTHORRE = re.compile (r'<font size=\+\d>Posted by (.*) on (.*? \d+, \d+ at \d+:\d+:\d+):')
	PARENTRE = re.compile (r'In Reply to: <a href="(\d+).html">(.*?)</a> posted by (.*) on (.*? \d+, \d+ at \d+:\d+:\d+):')
	AUTHORRE = re.compile (r'^(?:<a href="mailto:(?P<mail>.*?)">)?(?P<author>.*?)(?:</a>)?$')

	# for fromHtmlIndex()
	# finds data from wwwboard-indexfiles, as well as followups from message files
	INFORE = re.compile (r'^(?:<br>)?<!--top: (?P<id>\d+)--><li><a href="(?:messages/)?\d+.html">(?P<title>.*?)</a>\s*-?\s*<b>(?P<author>.*?)</b>\s*<i>(?P<date>.*?)</i>$')
	CHILDRENSTARTRE = re.compile (r'^<ul><!--insert: (\d+)-->$')
	CHILDRENENDRE = re.compile (r'</ul><!--end: (\d+)-->')

	def __init__ (self, id, title, written, author, parent, content):
		self.id = id
		self.title = title
		self.written = written
		self.author = author
		self.parent = parent
		self.content = content
	
	def __ne__ (x, y):
		return not (x == y)

	def __eq__ (x, y):
		# accept slightly different dates (+/- 1 minutes), maybe the server
		# time was not always synchronized...
		if y.written >= x.written-timedelta (minutes = 1) and y.written <= x.written+timedelta (minutes = 1):
			timeequal = True
		else:
			timeequal = False
		return x.id == y.id and x.title == y.title and timeequal and \
				x.author == y.author and x.parent == y.parent and x.content == y.content
	
	def __repr__ (self):
		return u'<Post #%i by %s: %s>' % (self.id, self.author, self.title)

	def merge (self, newpost, authorative):
		"""
		merge newpost into self

		@param newpost
		@param authorative overwrite existing data
		@param override_parent overwrite parent if not authorative
		"""

		if self.id != newpost.id:
			raise MustNotHappen ('wrong id, dude')

		# FIXME: ugly
		if authorative:
			if newpost.title is not None:
				self.title = newpost.title
			if newpost.written is not None:
				self.written = newpost.written
			if newpost.author is not None:
				self.author = newpost.author
			if newpost.parent is not None:
				self.parent = newpost.parent
			if newpost.content is not None:
				self.content = newpost.content
		else:
			if self.title is None:
				self.title = newpost.title
			if self.written is None:
				self.written = newpost.written
			if self.author is None:
				self.author = newpost.author
			if self.parent is None:
				self.parent = newpost.parent
			if self.content is None:
				self.content = newpost.content

	@classmethod
	def fromHtml (cls, file):
		txt = open (file, 'r').read ().decode ('latin1')
		try:
			id = int (os.path.basename (file).split ('.')[0])
		except ValueError:
			raise MustNotHappen ('unexpected filename ' + file)

		title = cls.TITLERE.findall (txt)
		if len (title) > 1:
			raise MustNotHappen ('more than one title found: ' + file)
		elif len (title) == 0:
			raise MustNotHappen ('no title found: ' + file)
		title = multireplace (entityDecode (title[0]), [(u'´', '\''), (u'`', '\'')])
		
		timeauthor = cls.TIMEAUTHORRE.findall (txt)
		if len (timeauthor) > 1:
			raise MustNotHappen ('more than one time/author found: ' + file)
		author, written = timeauthor[0]
		authormatch = cls.AUTHORRE.search (author)
		if authormatch:
			author = authormatch.groupdict ()
			if author['mail'] is not None:
				author = '%s <%s>' % (author['author'], author['mail'])
			else:
				author = author['author']
			# work around buggy board (y2k bug)
			written = written.replace (', 1920', ', 20')
			written = datetime.strptime (written, '%B %d, %Y at %H:%M:%S')
		else:
			raise MustNotHappen ('no author: ' + file)
		
		parentmatch = cls.PARENTRE.findall (txt)
		if parentmatch:
			if len (parentmatch) > 1:
				raise MustNotHappen ('more than one parent found: ' + file)
			parentmatch, = parentmatch
			if len (parentmatch) != 4:
				raise MustNotHappen ('regex broken')
			parent = int (parentmatch[0])
			parenttitle = parentmatch[1]
			parentauthor = parentmatch[2]
			# work around buggy board (y2k bug)
			parenttime = parentmatch[3].replace (', 1920', ', 20')
			parenttime = datetime.strptime (parenttime, '%B %d, %Y at %H:%M:%S')
		else:
			parent = cls.NOPARENT

		content = cls.CONTENTRE.findall (txt)
		if len (content) > 1:
			raise MustNotHappen ('more than one content part found: ' + file)
			return
		elif len (content) == 0:
			raise MustNotHappen ('no content found: ' + file)
		content = multireplace (content[0], [
				# html tags
				('<p>', '\n\n'),
				('<br>', '\n'),
				('<ul><font size=+1>', ''),
				('</font></ul>', ''),
				('<ul><font size=+0>', ''),
				# wrong apostrophs
				(u'´', '\''),
				(u'`', '\''),
				# strage character encodings (probably latin1)
				(u'\u0083G', '\''),
				(u'\u0081K', u'ß'),
				(u'\u0092', '\''),
				(u'\u0093', '"'),
				(u'\u0094', '"'),
				])
		content = entityDecode (content.strip ())

		if parent != cls.NOPARENT:
			parentpost = cls (parent, parenttitle, parenttime, parentauthor, None, None)
		else:
			parentpost = None
		# return tuple: this post, parent post (non-authorative), followups (non-authorative)
		return (cls (id, title, written, author, parent, content), parentpost, cls.fromHtmlIndex (file))
	
	@classmethod
	def fromHtmlIndex (cls, file):
		ret = []

		parent = [Post.NOPARENT]
		for line in open (file, 'r').read ().decode ('latin1').split ('\n'):
			line = line.strip ()
			infomatch = cls.INFORE.match (line)
			if infomatch:
				info = infomatch.groupdict ()
				try:
					written = datetime.strptime (info['date'], '%H:%M:%S %m/%d/%y')
				except ValueError:
					# work around buggy board software (y2k bug, prints 19XX as
					# two digit year, but 20XX as four digit year)
					try:
						written = datetime.strptime (info['date'], '%H:%M:%S %m/%d/%Y')
					except ValueError:
						# work around strange date errors by search&replace, probably y2k bug
						written = datetime.strptime (info['date'].replace ('/100', '/2000'), '%H:%M:%S %m/%d/%Y')
				ret.append (cls (int (info['id']), entityDecode (info['title']), written, entityDecode (info['author']), parent[-1], None))
			else:
				childrenstartmatch = cls.CHILDRENSTARTRE.match (line)
				if childrenstartmatch:
					id = int (childrenstartmatch.groups ()[0])
					parent.append (id)
				else:
					childrenendmatch = cls.CHILDRENENDRE.match (line)
					if childrenendmatch:
						id = int (childrenendmatch.groups ()[0])
						oldparent = parent.pop ()
						if oldparent != id:
							print file
							print line
							raise MustNotHappen ('parent ids messed up! %i != %i' % (oldparent, id))

		return ret

class Board(object):
	def __init__ (self):
		# posts tree with initial top-level dummy post
		self.posts = {}
		toppost = Post (Post.NOPARENT, '', datetime.now (), '', Post.NOPARENT, '')
		self.addPost (toppost)
	
	def addPost (self, p, authorative = False):
		"""
		add post to board, merge if post already exists

		@param p the new post
		@param authorative merge by replacing already available data (true) or
				merge by adding only the missing bits (false)
		"""
		if p is None:
			raise MustNotHappen ('no post')

		if p.id in self.posts:
			oldpost = self.posts[p.id]['post']
			self.posts[p.id] = None
			p.merge (oldpost, not authorative)

		self.posts[p.id] = {'post': p, 'children': []}

	def regenerateTree (self):
		"""
		regenerate children lists aka tree information. you have to call this
		method after adding all you fancy posts. calling it everytime you add a
		post is expensive.
		"""

		childsort = lambda x, y: cmp (x['post'].written, y['post'].written)

		for cid in self.posts.iterkeys ():
			if cid != Post.NOPARENT and self.posts[cid]['post'].parent is not None:
				# add post to parent's children list
				self.posts[self.posts[cid]['post'].parent]['children'].append (self.posts[cid])

		# sort child nodes by date
		for cid in self.posts.iterkeys ():
			self.posts[cid]['children'].sort (cmp=childsort)

	def _dumpRecursive (self, post, formatter, indent, dangling, banned = []):
		dumpedposts = [post['post'].id]

		if post['post'].id in banned:
			# ban all follow-ups too
			banned.extend ([child['post'].id for child in post['children']])
		else:
			formatter.walkPostStart (post['post'], indent, dangling)
			if post['post'].parent == Post.NOPARENT and indent != 0:
				raise MustNotHappen ('there\'s something wrong with the indent...')
			formatter.walkPostContent (post['post'], indent, dangling)

		for child in post['children']:
			dumpedposts.extend (self._dumpRecursive (child, formatter, indent+1, False, banned))

		if post['post'].id not in banned:
			formatter.walkPostEnd (post['post'], indent, dangling)

		return dumpedposts

	def dump (self, formatter, banned = []):
		formatter.walkStart ()

		dumpedposts = [Post.NOPARENT]
		for p in self.posts[Post.NOPARENT]['children']:
			dumpedposts.extend (self._dumpRecursive (p, formatter, 0, False, banned))

		# print dangling posts (parent missing)
		formatter.walkDanglingStart ()
		for id in self.posts.iterkeys ():
			if id not in dumpedposts:
				dumpedposts.extend (self._dumpRecursive (self.posts[id], formatter, 0, True, banned))
		formatter.walkDanglingEnd ()

		formatter.walkFinish ()

class Formatter (object):
	"""
	formatter template
	"""

	def walkStart (self):
		pass

	def walkPostStart (self, post, indent, dangling):
		pass
	
	def walkPostEnd (self, post, indent, dangling):
		pass
	
	def walkPostContent (self, post, indent, dangling):
		pass
	
	def walkFinish (self):
		pass
	
	def walkDanglingStart (self):
		pass
	
	def walkDanglingEnd (self):
		pass

class RawFormatter (Formatter):
	INDENTRE = re.compile (r'^', re.MULTILINE)

	def __init__ (self, outfd):
		self.outfd = outfd
	
	def walkPostContent (self, post, indent, dangling):
		postcontent = u"""
**********
id: %i
parent: %s
title: %s
author: %s
date: %s

%s
""" % (post.id, unicode (post.parent), post.title, post.author, post.written, post.content)

		self.outfd.write (self.INDENTRE.sub (indent*'\t', postcontent).encode ('utf-8'))
	
class HtmlFormatter (Formatter):
	def __init__ (self, outfd):
		self.outfd = outfd

	def cloakEmail (self, address):
		if address is not None:
			return address.replace ('@', '&#64;')
	
	def walkStart (self):
		self.outfd.write (u"""<html>
<head>
<title>noreo.com bboard</title>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<meta http-equiv="Content-Language" content="en" />
<style type="text/css">
div.post > div.post {
	margin-left: 1em;
}
.postcontent {
	padding: 0.5em;
	margin-bottom: 0.5em;
	border-left: 0.2em solid #000;
}
.incomplete > .postcontent {
	border-left: 0.2em solid #ff0;
}
.dangling > .postcontent {
	border-top: 0.2em solid #f00;
}
</style>
</head>
<body>
<h1>Noreo’s Webboard</h1>
<p>This is an archived version of Noreo’s Ocean Girl Webboard. Some parts are
unfortunately missing: Posts with a yellow border on the left do not have any
text, only the meta-data (author, title, date) have been preserved in these
cases. Posts with a red border at the top are missing their parent (context
missing).</p>
<h2>Posts</h2>
""".encode ('utf-8'))
	
	@staticmethod
	def _txt2html (txt):
		if txt is not None:
			txt = multireplace (txt, [('&', '&amp;'), ('"', '&quot;'), ('<', '&lt;'), ('>', '&gt;')])
			txt = txt.replace ('\n', '<br />\n')
		else:
			txt = ''
		return txt

	def walkPostStart (self, post, indent, dangling):
		moreclasses = ''
		if dangling:
			moreclasses += ' dangling'
		if post.content is None:
			moreclasses += ' incomplete'
		self.outfd.write ((u'<div id="post-%i" class="post%s">\n' % (post.id, moreclasses)).encode ('utf-8'))
	
	def walkPostEnd (self, post, indent, dangling):
		self.outfd.write ((u'</div><!-- end post %i -->\n' % post.id).encode ('utf-8'))
	
	def walkPostContent (self, post, indent, dangling):
		self.outfd.write ((u'<div class="postcontent">\n<p><a href="#post-%i">#%i</a>: <strong>%s</strong></p>\n<p>%s</p>\n<p>– %s @ %s</p>\n</div>\n' % (post.id, post.id, self._txt2html (post.title), self.cloakEmail (self._txt2html (post.content)), self.cloakEmail (self._txt2html (post.author)), (post.written))).encode ('utf-8'))
	
	def walkFinish (self):
		self.outfd.write (u'</body></html>'.encode ('utf-8'))
	
	def walkDanglingStart (self):
		self.outfd.write (u'<h2>Dangling posts</h2><p>Posts with missing parents</p>')
	
	def walkDanglingEnd (self):
		pass

if __name__ == '__main__':
	b = Board ()

	c = Counter ('import')
	for f in sys.stdin:
		f = f.strip ()

#		sys.stderr.write (f + '\n')
		if os.path.basename (f).startswith ('wwwboard') or os.path.basename (f).startswith ('wwwbaord'):
			posts = Post.fromHtmlIndex (f)
			# index files contain many errors, they're not authorative...
			for p in posts:
				b.addPost (p, False)
		else:
			p, parent, followups = Post.fromHtml (f)
			b.addPost (p, True)
			if parent is not None:
				b.addPost (parent, False)
			for p in followups:
				b.addPost (p, False)
	
	c = Counter ('regenerateTree')
	b.regenerateTree ()

	c = Counter ('export')
	# ban spam/off-topic messages
	banned = []
	if len (sys.argv) > 1:
		for e in open (sys.argv[1], 'r').readlines ():
			if not e.startswith ('#'):
				banned.append (int (e))
	b.dump (HtmlFormatter (sys.stdout), banned)

	c = Counter ('stats')
	incomplete = 0
	dangling = 0
	for p in b.posts.itervalues ():
		if p['post'].content is None:
			incomplete += 1
		if p['post'].parent not in b.posts:
			dangling += 1

	sys.stderr.write (unicode ({'posts': len (b.posts), 'incomplete': incomplete, 'dangling': dangling}) + '\n')

	c = None

